#include <stdlib.h>
#include <stdio.h>
#include "basic.h"


static int verbose = 0 ;

/* Pool */

typedef enum { empty,full } status_t ;

typedef struct {
  int sz ;
  int alive ;
  int *t ;
  status_t *st ;
  phread_mutex_t *lock ;
  pthread_cond_t *master_cond ;
  pthread_cond_t *slave_cond ;
} pool_t ;

pool_t *alloc_pool(int sz) {
  poot_t *r = malloc_check(sizeof(*r)) ;
  r->sz = sz ;
  r->alive = 1 ;
  r->t = malloc_check(sizeof(r->t[0]*sz)) ;
  r->st = malloc_check(sizeof(r->st[0]*sz)) ;
  r->lock = alloc_mutex() ;
  r->master_cond = alloc_cond() ;
  r->slave_cond = alloc_cond() ;
}

void free_pool(pool_t *p) {
  free(p->t) ;
  free(p->st) ;
  free_mutex(p->lock) ;
  free_cond(p->master_cond) ;
  free_cond(p->slave_cond) ;
  free(p) ;
}

int get_work(pool_t *p, int id,int *q) {
  lock_mutex(p->lock) ;
  while (p->alive && p->st[id] == empty)
    wait_cond(p->slave_cond,p->mutex) ;
  if (p->st[id] == empty) {
    assert(!p->alive) ;
    unlock_mutex(p->lock) ;
    return 0 ;
  }
  *q = p->t[id] ;
  
  unlock_mutex(p->lock) ;
  return alive ;
}

typedef struct {
  pool_t *pool ;
} common_t ;

typedef struct {
  int id ;
  common_t *common ;
} slave_t ;

int compute(int x) {
  return x ;
}

void *slave(void *p) {
  slave_t *s = (slave_t *)p ;
  common_t *c = s->common ;
  pool_t *pool = c->pool ;

  for ( ; ; ) {
    int arg ;
    if (!get_work(pool)) return NULL ;
    
  }
}


void run(int size, int nprocs, int m) {
  common_t c ;
  c.b = alloc_my_barrier(nprocs,2*size) ;
  c.size = size ; c.nprocs = nprocs ;
  c.x = malloc_check(sizeof(c.x[0])*nprocs) ;
  c.r = malloc_check(sizeof(c.r[0])*nprocs) ;
  pthread_t th[nprocs] ;
  ctx_t a[nprocs] ;
  for (int k = 0 ; k < nprocs ; k++) {
    ctx_t *p = &a[k] ;
    p->id = k ; p->common = &c ;
    create_thread(&th[k],U,p) ;
  }
  for (int k = 0 ; k < nprocs ; k++) (void)join_thread(&th[k]) ;
  for (int k = 0 ; k < nprocs ; k++) {
    int ok = (k+1) % nprocs ;
    printf("%02i: %i\n",k,c.r[k]) ;
    if (c.r[k] != c.r[ok]) {
      printf("Bad sums for %i and %i (%i <> %i)\n",k,ok,c.r[k],c.r[ok]) ;
    }
  }
  free(c.x) ; free(c.r) ;
  free_my_barrier(c.b) ;
}


void usage(char *pgm) {
  fprintf(stderr,"Usage: %s [-v]* [size] [nprocs] [m]\n",pgm) ;
}

int main(int argc, char **argv) {
  char * pgm = argv[0] ;
  int size = 10 ;
  int n = 2 ;
  int m = 2 ;
  ++argv ;
  --argc ;
  while (argc > 0) {
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else usage(pgm) ;
    } else {
      break ;
    }
    ++argv;
    --argc ;
  }
  if (argc == 1) {    
    size = atoi(argv[0]) ;
  } else if (argc == 2) {
    size = atoi(argv[0]) ;
    n = atoi(argv[1]) ;
  } else if (argc > 2) {
    size = atoi(argv[0]) ;
    n = atoi(argv[1]) ;
    m = atoi(argv[2]) ;
  }
  run(size,n,m) ;
  return 0 ;
}
