#include <stdlib.h>
#include <stdio.h>
#include "basic.h"
#include "stack.h"

/* Alloc/free */
stack_t *alloc_stack(int sz) {
  stack_t *p = malloc_check(sizeof(*p));
  /* Stack array */
  p->sz = sz ;
  p->sp = 0 ;
  p->t = calloc_check(sz,sizeof(p->t[0]));
  /* Kill */
  p->killed = 0 ;
  p->seen = 0 ;
  /* Lock and conditions */
  p->lock = alloc_mutex();
  p->is_empty = alloc_cond();
  p->is_full = alloc_cond();
  p->wait = alloc_cond();
  return p ;
}

void free_stack(stack_t *p) {
  free(p->t);
  free_mutex(p->lock);
  free_cond(p->is_empty);
  free_cond(p->is_full);
  free_cond(p->wait);
  free(p);
}

/* push/pop */

void push(stack_t *p, void *q) {
  lock_mutex(p->lock) ;
  while (p->sp >= p->sz) wait_cond(p->is_full,p->lock) ;
  int was_empty = p->sp <= 0 ;
  p->t[p->sp] = q ; p->sp++ ;
  if (was_empty) broadcast_cond(p->is_empty) ;
  unlock_mutex(p->lock) ;
}


void *pop(stack_t *p) {
  void *r ;
  lock_mutex(p->lock) ;
  while (p->sp <= 0) wait_cond(p->is_empty,p->lock) ;
  int was_full = p->sp >= p->sz ;
  p->sp-- ; r = p->t[p->sp] ;
  if (was_full) broadcast_cond(p->is_full) ;
  unlock_mutex(p->lock) ;
  return r ;
}

/* Kill */

void kill(stack_t *p,int nprocs) {
  // To be completed
}
